/* Les includes génériques de C */
#include <stdio.h>
#include <stdlib.h>

/* Les structures, protos et erreurs de la lib */
#include "libdtm.h"
 
#define COPY_VERTEX(ai,ii, aj,jj) {		\
      quad->ai ## x[ii] = quad->aj ## x[jj];	\
      quad->ai ## y[ii] = quad->aj ## y[jj];	\
      quad->ai ## z[ii] = quad->aj ## z[jj];	\
      quad->ai ## i[ii] = quad->aj ## i[jj];	\
      quad->ai ## d[ii] = quad->aj ## d[jj];	\
   }


/* Fonction that fuses the 3D points in the DTM, using the interpolation
   algorithm defined by Anthony. The code is therefore inspired (if not
   copied) from the fuse.c file from the lane library.
   */

/* --- Local functions ----------------------------------------------- */

static void	initCell(DTM_CELL *p, float z, unsigned char i, float c,
			float sigma);
static void	fusePixel(DTM_CELL *p1, DTM_CELL *p2, DTM_CELL *p);


/**********************************************************************
 *
 * Remplissage d'un MNT avec interpolation, avec éventuellement infos de
 * couleur. 
 *
 **********************************************************************/

int
dtm_fuseQuadIm3dIma1 (DATA_IM3D*	im3d,
		      T3D*		RobotRed_2_Start,
		      DTM*		mnt,
		      int		color,
		      unsigned char*	q_image,
		      int		nblig,
		      int		nbcol)

{
  int			nbligIm3d;
  int			nbcolIm3d;
  int			i, j;	/* Pour boucler sur les points de l'im3d */
  DATA_PT3D*		pt3d;	/* Un point de l'im3d */
  unsigned char*	pix = NULL;
  short			date;	/* La date (num. d'im3d) de mise à jour lors
				   de l'appel de cette fct */

  int			i_mnt;
  int			j_mnt;	/* Indices d'une cellule dans le MNT */
  DTM_CELL*		cell;
  int			nblig_mnt;
  int			nbcol_mnt;

  int			nb_pts3d_out; /* Les nombre de pts3d hors du MNT */

  DTM_CELL		new;
  DtmQuad *		quad = mnt->data;	/* Quad structure */
  DtmQuadSpan *		span;			/* Array of spans */

  int			l,c;			/* for fusion */
  int			lmin,ln;
  int			n;
  double		z,dz, lum,dlum;

  /* valididité des paramètres */
  if (mnt->cells == NIL_CELLS)
    return dtmError (ERR_DTM_CELLS_NON_ALLOUEES);

  if (im3d->header.points != POINTS_CALCULES)
    return dtmError (ERR_DTM_IM3D_NON_CALCULEE);

  if (im3d->header.coordonnees != COORDONNEES_CARTESIENNES)
    return dtmError (ERR_DTM_IM3D_COORDONNEES);

  /* On vérifie que l'im3d et l'image ont la même taille */
  nbligIm3d = im3d->header.nblig;
  nbcolIm3d = im3d->header.nbcol;
  if (color != 0) {
    if ( (nbligIm3d != nblig) || (nbcolIm3d != nbcol) ) {
      return dtmError (ERR_DTM_TAILLES_IMAGES);
    }
  }
  
  /* Slide the DTM */
  if (mnt->type == DTM_SLIDING) {
    dtm_slideDtm (mnt, RobotRed_2_Start);
  }

  /* Initialisations diverses et variées */
  date = mnt->nb_im3d + 1;
  nb_pts3d_out = 0;
  
  nblig_mnt = mnt->nblig;
  nbcol_mnt = mnt->nbcol;
  
  /* Boucle sur les points de l'im3d */
  pt3d = im3d->points_tab;
  if (color != 0)
    pix = q_image;
  switch (mnt->stockage) {
  case X_LIGNES:
     for (i = 0; i < nbligIm3d; i++) {
	for (j = 0; j < nbcolIm3d; j++, pt3d++, color?pix++:0) {

	   /* rotate buffered pixels (see quads.h) */
	   COPY_VERTEX(,0,,1);
	   COPY_VERTEX(,1,v,j);
	   COPY_VERTEX(,3,,2);
	   quad->d[2] = quad->vd[j] = 0;

	   if (pt3d->state != GOOD_PT3D) continue;

	   /* 3D point in mnt coordinates */
	   i_mnt = (pt3d->coord_1 - mnt->x_origine) / mnt->scale_x;
	   j_mnt = (pt3d->coord_2 - mnt->y_origine) / mnt->scale_y;
	  
	   if ((i_mnt < 0) || (i_mnt >= nblig_mnt)
	       || (j_mnt < 0) || (j_mnt >= nbcol_mnt))
	      nb_pts3d_out++;
	   else {

	      /* point is inside map */
	      quad->x[2] = quad->vx[j] = j_mnt; 
	      quad->y[2] = quad->vy[j] = i_mnt;
	      quad->z[2] = quad->vz[j] = pt3d->coord_3;
	      quad->i[2] = quad->vi[j] = color?*pix:0;
	      quad->d[2] = quad->vd[j] = 1;
	   }

	   /* quick test to figure out if we have something to fuse */
	   if (!*((int*)quad->d)) continue;

	   /* compute list of cells to fuse */
	   n = dtmMakeQuad(quad, &lmin, &ln);
	   if (!n) {
	      fputs("Empty quad !\n", stderr);
	      continue;
	   }

	   /* fusion (we're losing the sigma info here...) */
	   initCell(&new, pt3d->coord_3, color?*pix:0, 1.0/n, /*XXX*/0.0);

	   /* loop on the spans (i.e. lines) */
	   for(l=lmin, span=quad->span; l<lmin+ln; l++,span++) {

	      /* current elevation and luminace on this span */
	      z = span->zmin;
	      lum = span->imin;
	   
	      /* new cell to be fused (it has the confidence already) */
	      new.current_info.z_moyen = z;
	      new.current_info.luminance = lum;

	      /* old cell */
	      cell = dtm_one_cell(mnt, l, span->min);
	   
	      /* compute z increment along the span */
	      if (span->min == span->max)
		 dz = 0.0;
	      else
		 dz = (span->zmax - span->zmin)/(span->max - span->min);
	   
	      /* compute luminace increment along the span */
	      if (span->min == span->max)
		 dlum = 0.0;
	      else
		 dlum = (span->imax - span->imin)/(span->max - span->min);
	   
	      /* loop on the columns */
	      for(c=span->min; c<span->max || c==span->min; c++, cell++) {

		 fusePixel(cell, &new, cell);
		 cell->current_info.lastDate = date;

		 /* increment current values */
		 z+=dz;
		 lum+=dlum;
	     
		 /* update current new cell */
		 new.current_info.z_moyen = z;
		 new.current_info.luminance = lum;
	      }
	   }
	}
     }
     break;

     default:
	printf ("Mauvaise valeur du champ 'stockage' !!!\n");
	break;
  }
  
/*   printf ("%d points out\n", nb_pts3d_out); */

  mnt->cells = CELLS_DETERMINEES;
  if (color != 0)
    mnt->luminance = LUMINANCES_DETERMINEES;
  mnt->nb_im3d++;
  
  return (DTM_OK);
}


/*
 * --- initCell ---------------------------------------------------------
 *
 * Initialize a cell to reasonable default values;
 */

static void
initCell(DTM_CELL *p, float z, unsigned char i, float c, float sigma)
{
   p->state = FILLED;
   p->current_info.nb_points = 1;

   p->current_info.z_moyen = z;
   p->current_info.sigma_z = sigma;
   p->current_info.luminance = i;

   p->current_info.confidence = c;
}


/*
 * --- fusePixel --------------------------------------------------------
 *
 * Fuse two cells `p1' and `p2' into `p'.
 */

static void
fusePixel(DTM_CELL *p1, DTM_CELL *p2, DTM_CELL *p)
{
   double pr1, pr2, sum;

   pr1 = p1->current_info.confidence; 
   pr2 = p2->current_info.confidence; 
   sum = pr1 + pr2;

   if (sum == 0.0 || (p1->state == EMPTY && p2->state == EMPTY)) {
      p->state = EMPTY;
      p->current_info.confidence = 0.0;
      return;
   } 

   p->state = FILLED;
   p->current_info.nb_points = p1->current_info.nb_points + p2->current_info.nb_points;
   p->current_info.z_moyen = (pr1*p1->current_info.z_moyen + pr2*p2->current_info.z_moyen)/sum;
   p->current_info.sigma_z = (pr1*p1->current_info.sigma_z + pr2*p2->current_info.sigma_z)/sum;
   p->current_info.luminance = (pr1*p1->current_info.luminance + pr2*p2->current_info.luminance)/sum;
   p->current_info.confidence = sum;
}

