/**
 ** Copyright (c) 2000 LAAS/CNRS 
 **
 ** Anthony Mallet - Wed Nov 22 2000
 **
 ** $Log: pose.c,v $
 ** Revision 1.2  2006/12/01 16:05:23  simon
 ** Changes all "DEM" strings in "DTM".
 **
 ** Revision 1.1  2006/11/27 08:37:05  simon
 ** Autotoolization of the calife DTM library (Digital Elevation Map), now renamed
 ** 'dtm' (Digital Terrain Map).
 ** It depends on the calife openrobots library (t3d and 3dimages structures), but
 ** does not use the calife error system.
 ** It also depends on gdhe: there are functions that build a gdhe terrain structure
 ** (GDHE_MNT) from the internal DTM structure.
 **
 ** Revision 1.1  2004/12/08 10:45:19  simon
 ** Ajout de la librarie "dtm" (digital elevation model), qui était auparavent
 ** dans le calife de lama.
 **
 ** Revision 1.5  2001/09/13 10:47:14  mallet
 ** recuperation de la pose estimation de lane
 **
 ** Revision 1.4  2000/11/22 17:08:28  mallet
 ** calcul 6d
 **
 ** Revision 1.3  2000/11/22 16:30:02  mallet
 ** Positionne une erreur quand c'est nécessaire
 **
 ** Revision 1.2  2000/11/22 16:23:59  mallet
 ** Oups
 **
 ** Revision 1.1  2000/11/22 16:17:28  simon
 ** Ajout d'une fct d'Anthony pour affiner la position avant de fusionner une
 ** image 3D dans le mnt
 **
 **/
#ident "$Id: pose.c,v 1.2 2006/12/01 16:05:23 simon Exp $"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>


/* Les structures, protos et erreurs de la lib */
#include "libdtm.h"

#ifndef M_PI
#define M_PI            3.14159265358979323846
#endif

static void	amoeba(float **p, float *y, int ndim, float ftol,
		       float (*funk)(float *), int *nfunk);
static float	dtmDistance(float *y);


/*
 * dtmPoseRefinment -----------------------------------------------------
 *
 */

static DATA_IM3D *data3d;
static DTM *dtm;
static int imstep;

T3D *
dtmPoseRefinement(DATA_IM3D *im3d, DTM *mnt, int step)
{
   int i,j, k, jmax;
   float **p, y[8], max;
   int ntest;
   T3D* tpar;
   int di,dj;

   data3d = im3d;
   dtm = mnt;
   imstep = step;

   p = malloc(8 *sizeof(float*));
   for(i=1;i<=7;i++) p[i] = malloc(7*sizeof(float));

   /* P0 = 0 */
   for(i=1;i<=6;i++) p[1][i] = 0.0;

   for(k=0;k<2;k++) { /* iterate simplex twice */

      /* P1..6 = P0 +l.e1..6 */
      for(i=2;i<=7;i++) {
	 for(j=1;j<=6;j++) p[i][j] = p[1][j];

	 if (i<=4)
	    p[i][i-1] += 0.04;
	 else
	    p[i][i-1] += 0.05;
      }

      /* y */
      for(i=1;i<=7;i++) y[i] = dtmDistance(p[i]);

      /* mimimize */
      amoeba(p, y, 6, 1e-3, dtmDistance, &ntest);

      printf("%d evaluations\n", ntest);
      if (ntest == 0 && k==0) {
	printf ("No pixels");
	return NULL;
      }

      /* return transformation (evaluation only) */
      max=1e10;
      jmax = 1;
      for(j=1;j<=7;j++) if (y[j] < max) { max = y[j]; jmax = j; }
      for(j=1;j<=6;j++) p[1][j] = p[jmax][j];
   }

   for(di=1;di<=7;di++) {
      printf ("\t%.4f %d: ", y[di], di);
      for(dj=1;dj<=6;dj++) 
	 printf (" %d=%.4f", dj, p[di][dj]);
      printf ("\n");
   }

   tpar = t3dNew(T3D_EULER, T3D_ALLOW_CONVERSION);
   tpar->euler.euler[0] = p[1][1];
   tpar->euler.euler[1] = p[1][2];
   tpar->euler.euler[2] = p[1][3];
   tpar->euler.euler[3] = p[1][4];
   tpar->euler.euler[4] = p[1][5];
   tpar->euler.euler[5] = p[1][6];

   return tpar;
}

static float 
dtmDistance(float *params /* [1..6] */)
{
   DATA_PT3D *p;
   float x,y,z;
   short l, c, dl, dc, ok;
   int i, count, pcount;
   int width, height;
   float distance, onedistance, mindistance;
   T3D* tpar;
   DTM_CELL*		cell;

   /* transformation */
   tpar = t3dNew(T3D_EULER, T3D_ALLOW_CONVERSION);
   tpar->euler.euler[0] = params[1];
   tpar->euler.euler[1] = params[2];
   tpar->euler.euler[2] = params[3];
   tpar->euler.euler[3] = params[4];
   tpar->euler.euler[4] = params[5];
   tpar->euler.euler[5] = params[6];
   t3dConvertTo(T3D_MATRIX, tpar);

   /* distance */
   pcount = count = 0;
   distance = 0.0;

   width = data3d->header.nbcol;
   height = data3d->header.nblig;
   p = data3d->points_tab;

   for(i=0; i<width*height; i+=imstep, p+=imstep) if (p->state == GOOD_PT3D) {

      /* 3D point in local frame */
      x = 
	 p->coord_1 * tpar->matrix.matrix[0] +
	 p->coord_2 * tpar->matrix.matrix[1] +
	 p->coord_3 * tpar->matrix.matrix[2] +
	 tpar->matrix.matrix[3];
      y = 
	 p->coord_1 * tpar->matrix.matrix[4] +
	 p->coord_2 * tpar->matrix.matrix[5] +
	 p->coord_3 * tpar->matrix.matrix[6] +
	 tpar->matrix.matrix[7];
      z = 
	 p->coord_1 * tpar->matrix.matrix[8] +
	 p->coord_2 * tpar->matrix.matrix[9] +
	 p->coord_3 * tpar->matrix.matrix[10] +
	 tpar->matrix.matrix[11];

      /* cell */
      l = (x - dtm->x_origine) / dtm->scale_x;
      c = (y - dtm->y_origine) / dtm->scale_y;
	  
#define WINSIZE	1
      mindistance = 1e6;
      ok = 0;
      for(dl=-WINSIZE;dl<=WINSIZE;dl++) for(dc=-WINSIZE;dc<=WINSIZE;dc++) {
	 if ((l+dl < 0) || (l+dl >= dtm->nblig)
	  || (c+dc < 0) || (c+dc >= dtm->nbcol)) continue;

	 cell = dtm_one_cell (dtm, l, c);
	 if (cell->state == EMPTY) continue;

	 ok = 1;
	 onedistance = (z - cell->current_info.z_moyen) * (z - cell->current_info.z_moyen);

	 if (onedistance < mindistance) mindistance = onedistance;
      }
      if (!ok) continue;

      distance += mindistance;
      count++;
   }
   
   if (count == 0) return 1e6;

   return distance/count;
}


/*
 * simplex --------------------------------------------------------------
 *
 * cf Numerical Recipes
 */

#define TINY	1e-10
#define NMAX	500
#define GET_PSUM \
   for(j=1; j<=ndim; j++) {				\
      for(sum=0.0, i=1; i<=mpts; i++) sum += p[i][j];	\
      psum[j] = sum;					\
   }
#define SWAP(a,b) { swap=(a); (a)=(b); (b)=swap; }

static float	amotry(float **p, float *y, float *psum, int ndim,
			float (*funk)(float *), int ihi, float fac);

static void
amoeba(float **p, float *y, int ndim, float ftol,
       float (*funk)(float *), int *nfunk)
{
   int i, ihi, ilo, inhi, j, mpts=ndim+1;
   float rtol, sum, swap, ysave, ytry, *psum;

   psum = malloc(ndim * sizeof(float)); psum--;
   *nfunk = 0;
   GET_PSUM;

   for(;;) {
      ilo=1;
      
      ihi = y[1]>y[2] ? (inhi=2,1) : (inhi=1,2);
      for(i=1; i<=mpts; i++) {
	 if (y[i] <= y[ilo]) ilo=i;
	 if (y[i] > y[ihi]) {
	    inhi = ihi;
	    ihi = i;
	 } else if (y[i] > y[inhi] && i!=ihi) inhi=i;
      }
      rtol = 2.0*fabs(y[ihi]-y[ilo])/(fabs(y[ihi])+fabs(y[ilo])+TINY);
      
      if (rtol < ftol) {
	 SWAP(y[1], y[ilo]);
	 for(i=1; i<=ndim; i++) SWAP(p[1][i], p[ilo][i]);
	 break;
      }
      if (*nfunk >= NMAX) { fputs("NMAX exceeded\n", stderr); break; }
      *nfunk += 2;

      ytry = amotry(p,y,psum,ndim,funk,ihi,-1.0);
      if (ytry <= y[ilo])
	 ytry = amotry(p,y,psum,ndim,funk,ihi,2.0);
      else if (ytry >= y[inhi]) {
	 ysave = y[ihi];
	 ytry = amotry(p,y,psum,ndim,funk,ihi,0.5);
	 if (ytry >= ysave) {
	    for(i=1;i<=mpts;i++) {
	       if (i != ilo) {
		  for(j=1;j<=ndim;j++)
		     p[i][j]=psum[j]=0.5*(p[i][j]+p[ilo][j]);
		  y[i]=(*funk)(psum);
	       }
	    }
	    *nfunk += ndim;
	    GET_PSUM;
	 }
      } else --(*nfunk);
   }
   
   free(psum+1);   
}


static float
amotry(float **p, float *y, float *psum, int ndim,
       float (*funk)(float *), int ihi, float fac)
{
   int j;
   float fac1, fac2, ytry, *ptry;

   ptry = malloc(ndim * sizeof(float)); ptry--;
   fac1 = (1.0-fac)/ndim;
   fac2 = fac1-fac;
   for(j=1; j<=ndim; j++) ptry[j]=psum[j]*fac1-p[ihi][j]*fac2;
   ytry=(*funk)(ptry);
   if (ytry < y[ihi]) {
      y[ihi]=ytry;
      for(j=1; j<=ndim;j++) {
	 psum[j] += ptry[j]-p[ihi][j];
	 p[ihi][j]=ptry[j];
      }
   }
   
   free(ptry+1);
   return ytry;
}
