/* Les includes génériques de C */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "libdtm.h"

#ifndef MAXINT
#define MAXINT ((int) (((unsigned int) -1) / 2))
#endif

/**********************************************************************
 *
 * Reads a DTM from an USGS ASCII file.
 *
 **********************************************************************/

DTM*
dtm_readAsciiUsgs (FILE*	file,
		   int		xyStep,
		   double	scale,
		   double	zScale)

{
  int		i, j;		/* Indices of a DTM value */
  DTM*		mnt;
  DTM_CELL*	cell;
  int		nbLines, nbColumns;
  int		x, y;
  float		z;		/* Coordinates read in the file */
  int		xMax, yMax;
  int		xMin, yMin;	/* Max x and y values in the file */
  float		zMin;

  /* Creates the DTM */
  mnt = create_dtm ();
  if (mnt == NULL) {
    dtmError (ERR_DTM_MALLOC);
    return (NULL);
  }

  /* Loops on the coordinates to get xMax and yMax */ 
  xMax = - MAXINT;
  yMax = - MAXINT;
  xMin = MAXINT;
  yMin = MAXINT;
  zMin = MAXINT;
  while (fscanf (file, "%d %d %f\n", &x, &y, &z) == 3) {
    xMax = MAX (xMax, x);
    yMax = MAX (yMax, y);
    xMin = MIN (xMin, x);
    yMin = MIN (yMin, y);
    zMin = MIN (zMin, z);
  }

  /* Computes the DTM size */
  nbLines = (xMax - xMin) / xyStep;
  nbColumns = (yMax - yMin) / xyStep;

  /* Init the DTM */
  if (init_dtm_cells (mnt, xyStep * scale, xyStep * scale, nbLines, nbColumns,
		      0.0, 0.0, DTM_FIXED, X_LIGNES) != DTM_OK) {
      dtmError (ERR_DTM_MALLOC);
      return (NULL);
    }
  
  /* Read the first line the points */
  rewind (file);
  while (fscanf (file, "%d %d %f\n", &x, &y, &z) == 3) {
    i = (x - xMin) / xyStep;
    j = (y - yMin) / xyStep;
    cell = mnt->cells_tab + (i * nbColumns) + j;
    cell->state = FILLED;
    cell->current_info.nb_points = 1;
    cell->current_info.luminance = 0;
    cell->current_info.firstDate = 1;
    cell->current_info.lastDate = 1;
    cell->current_info.z_moyen = (z - zMin) * zScale;
    cell->current_info.sigma_z = 0.0;
  }
  
  /* Update some flags */
  mnt->cells = CELLS_DETERMINEES;

  return (mnt);
}


